;;;; -*- Mode: Lisp -*-

;;;; substitutions.lisp --
;;;; General CL structures unifier.
;;;; Substitution definitions.  Mostly a rehash of the usual SICP stuff.

(in-package "CL.EXT.DACF.UNIFICATION") ; DACF = Data And Control Flow.

;;;---------------------------------------------------------------------------
;;; Bindings.

(deftype binding () 'cons)

(deftype bindings () 'list) ; An A-LIST.

(defun make-binding (variable value)
  (cons variable value))

(defun extend-bindings (variable value bindings)
  (acons variable value bindings))


(defun binding-variable (b)
  (declare (type binding b))
  (car b))

(defun (setf binding-variable) (v b)
  (declare (type binding b))
  (setf (car b) v))


(defun binding-value (b)
  (declare (type binding b))
  (cdr b))


(defun (setf binding-value) (v b)
  (declare (type binding b))
  (setf (cdr b) v))



(define-condition unification-variable-unbound (unbound-variable)
  ()
  )

(define-condition unification-failure (simple-error)
  ())


;;;---------------------------------------------------------------------------
;;; Frames.

(defstruct frame
  (bindings () :type bindings))

(defun empty-frame-p (f)
  (declare (type frame f))
  (null (frame-bindings f)))


(defun find-variable-binding-in-frame (v f)
  (declare (type frame f))
  (assoc v (frame-bindings f)))


(defun find-variable-value-in-frame (v f)
  (declare (type frame f))
  (let ((b (find-variable-binding-in-frame v f)))
    (declare (type (or null binding) b))
    (if b
        (values (cdr b) t)
        (values nil nil))))


;;;---------------------------------------------------------------------------
;;; Environments.

(defstruct (environment (:print-object print-environment)
                        (:copier nil))
  (frames () :type list))

(defun print-environment (env stream)
  (if *print-readably*
      (format stream "#S(ENVIRONMENT FRAMES ~S)"
              (environment-frames env))
      (print-unreadable-object (env stream :type nil :identity t)
        (format stream "~:[~;EMPTY ~]UNIFY ENVIRONMENT: ~D frame~:P"
                (empty-environment-p env)
                (list-length (environment-frames env))))))

(deftype substitution () 'environment)

(defun substitution-p (x) (environment-p x))

(defun first-frame (env)
  (first (environment-frames env)))


(defun make-empty-environment ()
  (make-environment :frames (list (make-frame))))

(defun copy-environment (env)
  (make-environment :frames (copy-list (environment-frames env))))

(defun make-shared-environment (env)
  (make-environment :frames (environment-frames env)))

(defun empty-environment-p (env &aux (env-frames (environment-frames env)))
  (declare (type environment env))
  (and (= 1 (list-length env-frames))
       (empty-frame-p (first env-frames))))

(defparameter *null-environment* (make-empty-environment))


(defun find-variable-value (variable &optional (env *null-environment*) errorp)
  (declare (type environment env))
  (labels ((find-var-value (frames)
             (cond (frames
                    (multiple-value-bind (val foundp)
                        (find-variable-value-in-frame variable (first frames))
                      (if foundp
                          (values val t)
                          (find-var-value (rest frames)))))
                   (errorp
                    (error 'unification-variable-unbound :variable variable))
                   (t (values nil nil))))
           )
    (find-var-value (environment-frames env))))



(defun extend-environment (var pat env)
  (let ((first-frame (first-frame env)))
    (setf (frame-bindings first-frame)
          (extend-bindings var pat (frame-bindings first-frame)))
    env))


;;;; end of file -- substitutions.lisp --
