;;;; -*- Mode: Lisp -*-

;;;; template-componds.lisp


(in-package "CL.EXT.DACF.UNIFICATION") ; DACF = Data And Control Flow.

;;;; Compound templates for unions and intersections.
;;;; This should be merged in 'template-hierachy.lisp'.  It depends on it.


;;;; Templates.
;;;; Essentially we extend the type specifier language.
;;;; The interesting thing is that we need to specify how a "match"
;;;; between a template and a CL object is performed.

;;;; A template can also be
;;;;
;;;;    <template> ::= #| templates from template-hierarchy.lisp |#
;;;;               |   <union template>
;;;;               |   <conjunction template>
;;;;               |   <negation template>

;;;; Note:  I am not sure this is the right way to do this, but it is
;;;; easy to add.


(defclass compound-template (template) ())

(defgeneric compound-template-p (x)
  (:method ((x compound-template)) t)
  (:method ((x t)) nil))


(defclass union-template (compound-template) ())

(defgeneric union-template-p (x)
  (:method ((x union-template)) t)
  (:method ((x t)) nil))


(defclass conjunction-template (compound-template) ())

(defgeneric conjunction-template-p (x)
  (:method ((x conjunction-template)) t)
  (:method ((x t)) nil))


(defclass negation-template (compound-template) ())

(defgeneric negation-template-p (x)
  (:method ((x negation-template)) t)
  (:method ((x t)) nil))



(defmethod make-template ((kind (eql 'or)) (spec cons))
  (make-instance 'union-template :spec spec))

(defmethod make-template ((kind (eql 'and)) (spec cons))
  (make-instance 'conjunction-template :spec spec))

(defmethod make-template ((kind (eql 'not)) (spec cons))
  (if (/= (list-length spec) 2)
      (error 'unification-template-error
             :format-control "Incorrect syntax for negation template ~S."
             :format-arguments (list spec))
      (make-instance 'negation-template :spec spec)))




;;; Implementation.

;;; Compound templates.
;;; Specification is
;;;
;;;    (<op> <template>*)

(defun compound-template-compounds (x)
  (declare (type compound-template x))
  (assert (compound-template-p x) (x) "Non compound template ~S." x)
  (rest (template-spec x)))


;;; Unification.

(defmethod unify ((ct1 compound-template) (ct2 compound-template)
                  &optional (env (make-empty-environment)))
  (error 'unification-failure
         :format-control "Do not know how unify the two compound templates: ~S ~S."
         :format-arguments (list ct1 ct2)))



(defmethod unify ((ut union-template) (o t)
                  &optional (env (make-empty-environment)))
  (let ((ts (compound-template-compounds ut)))
    (if ts
        (loop for tc in ts
              for subst = (ignore-errors (unify tc o env))
              when subst return subst
              finally
              (error 'unification-failure
                     :format-control "Cannot unify ~S with the disjunction template ~S ."
                     :format-arguments (list o ut)))
        (error 'unification-failure
               :format-control "Cannot unify ~S with the empty disjunction template."
               :format-arguments (list o)))))


(defmethod unify ((ct conjunction-template) (o t)
                  &optional (env (make-empty-environment)))
  (let ((ts (compound-template-compounds ct)))
    (if ts
        (loop for tc in ts
              for subst = (ignore-errors (unify tc o env))
              then (ignore-errors (unify tc o subst))
              when (not subst)
              do (error 'unification-failure
                        :format-control "Cannot unify ~S with the conjucntion template ~S."
                        :format-arguments (list o ct))
              finally (return subst))
        env ; This may be wrong.  Maybe an environment with the
            ; variables in O is in order. 
        )))


(defmethod unify ((nt negation-template) (o t)
                  &optional (env (make-empty-environment)))
  (handler-case
      (unify (first (compound-template-compounds nt)) o env)
    (unification-failure () env)
    (error (e) (error e))
    (:no-error (&rest unification-results)
      (declare (ignore unification-results))
      (error 'unification-failure
             :format-control "Cannot unify ~S with the negation template ~S."
             :format-arguments (list o nt)))))

  

;;; end of file -- templates.lisp --
